package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntSet
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Player
import net.minecraft.world.entity.player.StackedContents
import net.minecraft.world.inventory.StackedContentsCompatible
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import net.minecraft.world.item.crafting.RecipeInput
import ru.dbotthepony.mc.otm.core.collect.any
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.isNotEmpty
import java.util.function.Predicate
import java.util.stream.Stream
import java.util.stream.StreamSupport

/**
 * "Backward-compatible" enhanced container interface, where all methods can be derived/emulated from existing [IContainer] ([Container]) code
 *
 * This is useful because it allows to interact with actually enhanced and regular containers through unified interface,
 * and actual implementations of this interface are likely to provide efficient method implementations in place of derived/emulated ones.
 */
interface IEnhancedContainer : IContainer, RecipeInput, Iterable<ItemStack>, StackedContentsCompatible {
	// provide non-ambiguous get and set operators
	operator fun get(slot: Int): ItemStack {
		return getItem(slot)
	}

	operator fun set(slot: Int, value: ItemStack) {
		setItem(slot, value)
	}

	fun containerSlot(slot: Int): IContainerSlot {
		return IContainerSlot.Simple(slot, this)
	}

	override fun fillStackedContents(contents: StackedContents) {
		forEach { contents.accountStack(it) }
	}

	/**
	 * Returns iterator over **all** slots this container has
	 */
	fun slotIterator(): Iterator<IContainerSlot> {
		return (0 until containerSize).iterator().map { containerSlot(it) }
	}

	fun nonEmptySlotIterator(): Iterator<IContainerSlot> {
		return slotIterator().filter { it.isNotEmpty }
	}

	private fun slotIterator(allowedSlots: IntCollection, predicate: Predicate<ItemStack>): IntIterator {
		return object : IntIterator() {
			private val parent = allowedSlots.intIterator()
			private var foundNext = false
			private var next = -1

			private fun findNext() {
				if (!foundNext) {
					foundNext = true
					next = -1

					while (parent.hasNext()) {
						val i = parent.nextInt()

						if (predicate.test(this@IEnhancedContainer[i])) {
							next = i
							break
						}
					}
				}
			}

			override fun nextInt(): Int {
				findNext()

				if (next == -1)
					throw NoSuchElementException()

				foundNext = false
				val next = next
				this.next = -1
				return next
			}

			override fun hasNext(): Boolean {
				findNext()
				return next != -1
			}
		}
	}

	fun emptySlotIndexIterator(): IntIterator {
		return emptySlotIndexIterator(slotRange)
	}

	fun nonEmptySlotIndexIterator(): IntIterator {
		return nonEmptySlotIndexIterator(slotRange)
	}

	fun emptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		return slotIterator(allowedSlots) { it.isEmpty }
	}

	fun nonEmptySlotIndexIterator(allowedSlots: IntCollection): IntIterator {
		return slotIterator(allowedSlots) { it.isNotEmpty }
	}

	fun slotWithItemIterator(item: Item): IntIterator {
		return slotWithItemIterator(item, slotRange)
	}

	fun slotWithItemIterator(item: Item, allowedSlots: IntCollection): IntIterator {
		return slotIterator(allowedSlots) { it.isNotEmpty && it.item === item }
	}

	fun nextEmptySlot(startIndex: Int): Int {
		for (i in startIndex until containerSize) {
			if (this[i].isEmpty) {
				return i
			}
		}

		return -1
	}

	fun nextNonEmptySlot(startIndex: Int): Int {
		for (i in startIndex until containerSize) {
			if (this[i].isNotEmpty) {
				return i
			}
		}

		return -1
	}

	fun nextSlotWithItem(startIndex: Int, item: Item): Int {
		var find = nextNonEmptySlot(startIndex)

		while (find != -1 && this[find].item !== item)
			find = nextNonEmptySlot(find + 1)

		return find
	}

	fun setChanged(slot: Int) {
		return setChanged()
	}

	fun getMaxStackSize(slot: Int, itemStack: ItemStack): Int {
		return maxStackSize
	}

	/**
	 * Returns iterator over **non-empty** [ItemStack]s inside this container
	 */
	override fun iterator(): Iterator<ItemStack> {
		return (0 until containerSize).iterator().map { this[it] }.filter { it.isNotEmpty }
	}

	override fun isEmpty(): Boolean {
		return super.isEmpty()
	}

	val hasEmptySlots: Boolean get() {
		for (i in 0 until containerSize) {
			if (this[i].isEmpty) {
				return true
			}
		}

		return false
	}

	override fun size(): Int {
		return containerSize
	}

	override fun countItem(item: Item): Int {
		var count = 0

		for (stack in this) {
			if (stack.item === item) {
				count += stack.count
			}
		}

		return count
	}

	override fun hasAnyOf(items: Set<Item>): Boolean {
		if (Items.AIR in items && hasEmptySlots)
			return true

		return iterator().any { it.item in items }
	}

	override fun hasAnyMatching(predicate: Predicate<ItemStack>): Boolean {
		if (predicate.test(ItemStack.EMPTY) && hasEmptySlots)
			return true

		return iterator().any(predicate)
	}

	fun toList(): MutableList<ItemStack> {
		val list = ArrayList<ItemStack>(containerSize)

		for (i in 0 until containerSize) {
			list.add(this[i])
		}

		return list
	}

	fun addItem(stack: ItemStack, simulate: Boolean, slots: IntCollection = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null): ItemStack {
		if (stack.isEmpty || slots.isEmpty())
			return stack

		val copy = stack.copy()

		// двигаем в одинаковые слоты
		for (slot in slotWithItemIterator(stack.item, slots)) {
			if (ItemStack.isSameItemSameComponents(this[slot], copy)) {
				val slotStack = this[slot]
				val slotLimit = getMaxStackSize(slot, slotStack)

				if (slotStack.count < slotLimit) {
					val newCount = (slotStack.count + copy.count).coerceAtMost(slotLimit)
					val diff = newCount - slotStack.count

					if (!simulate) {
						slotStack.count = newCount
						setChanged(slot)

						if (popTime != null) {
							slotStack.popTime = popTime
						}
					}

					copy.shrink(diff)

					if (copy.isEmpty) {
						return ItemStack.EMPTY
					}
				}
			}
		}

		if (!onlyIntoExisting) {
			// двигаем в пустые слоты
			for (slot in emptySlotIndexIterator(slots)) {
				val diff = copy.count.coerceAtMost(getMaxStackSize(slot, stack))

				if (!simulate) {
					val copyToPut = copy.copy()
					copyToPut.count = diff
					this[slot] = copyToPut
					setChanged()
				}

				copy.shrink(diff)

				if (copy.isEmpty)
					return ItemStack.EMPTY
			}
		}

		return copy
	}

	/**
	 * Unlike [addItem], modifies original [stack]
	 *
	 * @return Whenever [stack] was modified
	 */
	fun consumeItem(stack: ItemStack, simulate: Boolean, slots: IntCollection = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null): Boolean {
		if (stack.isEmpty || slots.isEmpty())
			return false

		val result = addItem(stack, simulate, slots, onlyIntoExisting, popTime)
		if (!simulate) stack.count = result.count
		return result.count != stack.count
	}

	fun fullyAddItem(stack: ItemStack, slots: IntCollection = slotRange, onlyIntoExisting: Boolean = false, popTime: Int? = null): Boolean {
		if (!addItem(stack, true, slots, onlyIntoExisting, popTime).isEmpty)
			return false

		return addItem(stack, false, slots, onlyIntoExisting, popTime).isEmpty
	}

	fun stream(): Stream<ItemStack> {
		return StreamSupport.stream(spliterator(), false)
	}

	private class Wrapper(private val parent: Container) : IEnhancedContainer {
		override fun clearContent() {
			return parent.clearContent()
		}

		override fun setChanged() {
			return parent.setChanged()
		}

		override fun getContainerSize(): Int {
			return parent.containerSize
		}

		override fun getItem(slot: Int): ItemStack {
			return parent.getItem(slot)
		}

		override fun removeItem(slot: Int, amount: Int): ItemStack {
			return parent.removeItem(slot, amount)
		}

		override fun removeItemNoUpdate(slot: Int): ItemStack {
			return parent.removeItemNoUpdate(slot)
		}

		override fun setItem(slot: Int, itemStack: ItemStack) {
			return parent.setItem(slot, itemStack)
		}

		override fun stillValid(player: Player): Boolean {
			return parent.stillValid(player)
		}

		override fun getMaxStackSize(): Int {
			return parent.maxStackSize
		}

		override fun startOpen(player: Player) {
			parent.startOpen(player)
		}

		override fun stopOpen(player: Player) {
			parent.stopOpen(player)
		}

		override fun canPlaceItem(slot: Int, itemStack: ItemStack): Boolean {
			return parent.canPlaceItem(slot, itemStack)
		}

		override fun canTakeItem(container: Container, slot: Int, itemStack: ItemStack): Boolean {
			return parent.canTakeItem(container, slot, itemStack)
		}

		override fun isEmpty(): Boolean {
			return parent.isEmpty
		}

		override fun hasAnyMatching(predicate: Predicate<ItemStack>): Boolean {
			return parent.hasAnyMatching(predicate)
		}

		override fun hasAnyOf(items: Set<Item>): Boolean {
			return parent.hasAnyOf(items)
		}

		override fun countItem(item: Item): Int {
			return parent.countItem(item)
		}
	}

	companion object {
		fun wrap(other: Container): IEnhancedContainer {
			if (other is IEnhancedContainer)
				return other

			return Wrapper(other)
		}
	}
}
